/*
 * Copyright (c) 1999-2000, Image Power, Inc. and the University of
 *   British Columbia.
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved. 
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * Command Line Option Parsing Library
 *
 * $Id$
 */

/******************************************************************************\
* Includes.
\******************************************************************************/

#include <stdio.h>
#include <string.h>
#include "getoptx.h"
#include "memalloc.h"

/******************************************************************************\
* Macros.
\******************************************************************************/

#ifndef min
#define	min(x, y)	(((x) < (y)) ? (x) : (y))
#endif

/******************************************************************************\
* Global data.
\******************************************************************************/

int optind = 0;
int opterr = 1;
char *optarg = 0;

/******************************************************************************\
* Code.
\******************************************************************************/

static opt_t *optlookup(opt_t *opts, char *name)
{
	opt_t *opt;

	for (opt = opts; opt->id >= 0 && opt->name; ++opt) {
		if (!strcmp(opt->name, name)) {
			return opt;
		}
	}
	return 0;
}

int getoptx(int argc, char **argv, opt_t *opts)
{
	char *cp;
	int id;
	int hasarg;
	opt_t *opt;
	char *s;

	if (!optind) {
		optind = min(1, argc);
	}
	while (optind < argc) {
		s = cp = argv[optind];
		if (*cp == '-') {
			/* We are processing an option. */
			++optind;
			if (*++cp == '-') {
				/* We are processing a long option. */
				++cp;
				if (*cp == '\0') {
					/* This is the end of the options. */
					return GETOPT_EOF;
				}
				if (!(opt = optlookup(opts, cp))) {
					if (opterr) {
						fprintf(stderr, "unknown option %s\n", s);
					}
					return GETOPT_ERR;
				}
				hasarg = (opt->flags & OPT_HASARG) != 0;
				id = opt->id;
			} else {
				/* We are processing a short option. */
				if (strlen(cp) != 1 ||
				  !(opt = optlookup(opts, cp))) {
					if (opterr) {
						fprintf(stderr, "unknown option %s\n", s);
					}
					return GETOPT_ERR;
				}
				hasarg = (opt->flags & OPT_HASARG) != 0;
				id = opt->id;
			}
			if (hasarg) {
				/* The option has an argument. */
				if (optind >= argc) {
					if (opterr) {
						fprintf(stderr, "missing argument for option %s\n", s);
					}
					return GETOPT_ERR;
				}
				optarg = argv[optind];
				++optind;
			} else {
				/* The option does not have an argument. */
				optarg = 0;
			}
			return id;
		} else {
			/* We are not processing an option. */
			return GETOPT_EOF;
		}
	}
	return GETOPT_EOF;
}
