/*
 * Copyright (c) 1999-2000 Image Power, Inc. and the University of
 *   British Columbia.
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * $Id$
 */

#ifndef JPCDEC_H
#define JPCDEC_H

/******************************************************************************\
* Includes.
\******************************************************************************/

#include "tsfb.h"
#include "stream.h"
#include "bs.h"
#include "tagtree.h"
#include "cs.h"
#include "jpccod.h"
#include "mqdec.h"
#include "t2cod.h"

/******************************************************************************\
* Constants.
\******************************************************************************/

/******************************************************************************\
* Types.
\******************************************************************************/

typedef struct {
	uint_fast16_t ind;
	uint_fast32_t len;
	unsigned char *data;
} ppxstabent_t;

typedef struct {
	int numents;
	int maxents;
	ppxstabent_t **ents;
} ppxstab_t;

typedef struct {
	int numstreams;
	int maxstreams;
	stream_t **streams;
} streamlist_t;

typedef struct {

	int flags;

	uint_fast8_t csty;
	/* Flag for explicitly specified precinct sizes. */

	uint_fast8_t numrlvls;

	uint_fast8_t cblkwidthexpn;

	uint_fast8_t cblkheightexpn;

	uint_fast8_t qmfbid;

	uint_fast8_t qsty;

	uint_fast16_t numstepsizes;

	uint_fast16_t stepsizes[3 * JPC_MAXRLVLS + 1];

	uint_fast8_t numguardbits;

	uint_fast8_t roishift;

	uint_fast8_t cblkctx;

	uint_fast8_t prcwidthexpns[JPC_MAXRLVLS];

	uint_fast8_t prcheightexpns[JPC_MAXRLVLS];

} jpd_ccp_t;

typedef struct {

	int flags;

	pchglist_t *pchglist;

	uint_fast8_t prgord;

	uint_fast16_t numlyrs;

	uint_fast8_t mctid;

	uint_fast8_t csty;
	/* SOP, EPH */

	uint_fast16_t numcomps;

	jpd_ccp_t *ccps;

} jpd_cp_t;

typedef struct jpd_seg_ {

	struct jpd_seg_ *next;
	/* The next segment in the list. */

	struct jpd_seg_ *prev;
	/* The previous segment in the list. */

	int passno;
	/* The starting pass number for this segment. */

	int numpasses;
	/* The number of passes in this segment. */

	int maxpasses;
	/* The maximum number of passes in this segment. */

	int type;
	/* The type of data in this segment. */

	stream_t *stream;

	int cnt;

	int complete;
	/* Flag indicating if the segment has been terminated. */

	int lyrno;

} jpd_seg_t;

typedef struct {
	jpd_seg_t *head;
	jpd_seg_t *tail;
} jpd_seglist_t;

typedef struct {

	int numpasses;
	/* The number of passes. */

	jpd_seglist_t segs;
	/* A list of segments that still need to be decoded. */

	jpd_seg_t *curseg;
	/* The first incomplete/partial segment. */

	int numimsbs;

	int numlenbits;
	/* The number of bits used to encode pass data lengths. */

	int firstpassno;
	mqdec_t *mqdec;
	bitstream_t *nulldec;
	matrix_t *flags;
	matrix_t *data;

} jpd_cblk_t;

typedef struct {

	uint_fast32_t xstart;
	uint_fast32_t ystart;
	uint_fast32_t xend;
	uint_fast32_t yend;
	int numhcblks;
	int numvcblks;
	int numcblks;

	jpd_cblk_t *cblks;
	/* The per code block information. */

	tagtree_t *incltagtree;
	/* The inclusion tag tree. */

	tagtree_t *numimsbstagtree;
	/* The insignificant MSBs tag tree. */

} jpd_prc_t;

typedef struct jpd_band_tag {

	jpd_prc_t *prcs;

	matrix_t *data;
	/* The band data. */

	int orient;
	/* The orientation of this band (i.e., LL, LH, HL, or HH). */

	int stepsize;
	/* The quantization step size. */

	int numbps;
	/* The number of bit planes for this band. */

	fix_t absstepsize;

	int analgain;

int roishift;

} jpd_band_t;

typedef struct {

	int numbands;
	/* The number of bands associated with this resolution. */

	jpd_band_t *bands;
	/* The per-band information. */

	uint_fast32_t xstart;
	/* The x-coordinate of the top-left corner of the tile-component
	  at this resolution. */

	uint_fast32_t ystart;
	/* The y-coordinate of the top-left corner of the tile-component
	  at this resolution. */

	uint_fast32_t xend;
	/* The x-coordinate of the bottom-right corner of the tile-component
	  at this resolution. */

	uint_fast32_t yend;
	/* The y-coordinate of the bottom-right corner of the tile-component
	  at this resolution. */

	int prcwidthexpn;
	/* The exponent value for the nominal precinct width measured
	  relative to the associated LL band. */

	int prcheightexpn;
	/* The exponent value for the nominal precinct height measured
	  relative to the associated LL band. */

	int numhprcs;
	/* The number of precincts in the horizontal direction. */

	int numvprcs;
	/* The number of precincts in the vertical direction. */

	int numprcs;
	/* The number of precincts. */

	int cbgwidthexpn;
	/* The exponent value for the nominal code block group width.
	  This quantity is associated with the next lower resolution level
	  (assuming that there is one). */

	int cbgheightexpn;
	/* The exponent value for the nominal code block group height
	  This quantity is associated with the next lower resolution level
	  (assuming that there is one). */

	uint_fast16_t cblkwidthexpn;
	uint_fast16_t cblkheightexpn;

} jpd_rlvl_t;

typedef struct {

	uint_fast32_t xstart;

	uint_fast32_t ystart;

	uint_fast32_t xend;

	uint_fast32_t yend;

	matrix_t *data;
	/* The component data for the current tile. */

	uint_fast16_t numrlvls;
	/* The number of resolution levels. */

	jpd_rlvl_t *rlvls;
	/* The per resolution level information. */

	tsfb_t *tsfb;

} jpd_tcomp_t;

/* tile states. */
#define	JPD_TILE_INIT	0
#define	JPD_TILE_ACTIVE	1
#define	JPD_TILE_ACTIVELAST	2
#define	JPD_TILE_DONE	3

typedef struct {

	int state;
	uint_fast32_t xstart;
	uint_fast32_t ystart;
	uint_fast32_t xend;
	uint_fast32_t yend;

	ppxstab_t *pptstab;
	stream_t *pkthdrstream;
	long pkthdrstreampos;

	jpd_cp_t *cp;
	/* The coding parameters for this tile. */

	jpd_tcomp_t *tcomps;

	int partno;
	/* The next expected tile-part number. */

	int numparts;
	/* The number of tile-parts. */

	int realmode;

	pi_t *pi;

} jpd_tile_t;

typedef struct {

	uint_fast32_t hstep;
	uint_fast32_t vstep;

	uint_fast32_t width;

	uint_fast32_t height;

	uint_fast16_t prec;
	/* The precision of the component sample data. */

	bool sgnd;
	/* The format of the component sample data. */
	/* This is currently unused and should be set to zero. */

	uint_fast32_t hsubstep;
	uint_fast32_t vsubstep;

} dec_cmpt_t;

/* Decoder state information. */

typedef struct {

	image_t *image;
	/* The decoded image. */

	uint_fast32_t xstart;
	/* The horizontal offset from the origin of the reference grid to the
	  left side of the image area. */

	uint_fast32_t ystart;
	/* The vertical offset from the origin of the reference grid to the
	  top side of the image area. */

	uint_fast32_t xend;
	uint_fast32_t yend;

	uint_fast32_t tilewidth;
	/* The nominal tile width in units of the image reference grid. */

	uint_fast32_t tileheight;
	/* The nominal tile height in units of the image reference grid. */

	uint_fast32_t tilexoff;
	/* The horizontal offset from the origin of the reference grid to the
	  left side of the first tile. */

	uint_fast32_t tileyoff;
	/* The vertical offset from the origin of the reference grid to the
	  top side of the first tile. */

	int numhtiles;
	/* The number of tiles in the vertical direction. */

	int numvtiles;
	/* The number of tiles in the horizontal direction. */

	int numtiles;
	/* The number of tiles. */

	jpd_tile_t *tiles;
	/* The per-tile information. */

	jpd_tile_t *curtile;
	/* The tile currently being processed. */

	int numcomps;
	/* The number of components. */

	stream_t *in;
	/* The JP2 input stream. */

	jpd_cp_t *cp;
	/* The default coding parameters for all tiles. */

	int maxlyrs;

	int ppmseqno;
	/* The next expected PPM marker segment sequence number. */

	int state;
	/* The current state for code stream processing. */

	dec_cmpt_t *cmpts;

	ppxstab_t *ppmstab;
	streamlist_t *pkthdrstreams;

	long curtileendoff;

jpc_cstate_t *cstate;
/* This is required by T2 decoder. */

} jpd_t;

typedef struct {

	int dbglevel;

	int maxlyrs;

	int reserved;

} jp_importopts_t;

/******************************************************************************\
* Functions.
\******************************************************************************/

jpd_seg_t *seg_alloc();
void seglist_remove(jpd_seglist_t *list, jpd_seg_t *node);
void seg_destroy(jpd_seg_t *seg);
void jpd_setdbglvl(int dbglvl);
int jpd_getdbglvl();
void seglist_insert(jpd_seglist_t *list, jpd_seg_t *ins, jpd_seg_t *node);

#endif
