/*
 * Copyright (c) 1999-2000 Image Power, Inc. and the University of
 *   British Columbia.
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * MQ Arithmetic Encoder
 *
 * $Id$
 */

#ifndef MQENC_H
#define MQENC_H

/******************************************************************************\
* Includes.
\******************************************************************************/

#include "types.h"
#include "stream.h"
#include "mqcod.h"

#define	MQENC_DEFTERM	0	/* default termination */
#define	MQENC_PTERM		1		/* predictable termination */

/******************************************************************************\
* Types.
\******************************************************************************/

/* MQ arithmetic encoder. */
typedef struct {

	uint_fast32_t creg;
	/* The C register. */

	uint_fast32_t areg;
	/* The A register. */

	uint_fast32_t ctreg;
	/* The CT register. */

	int maxctxs;
	/* The maximum number of contexts. */

	mqstate_t **ctxs;
	/* The per-context information. */

	mqstate_t **curctx;
	/* The current context. */

	stream_t *out;
	/* The stream for encoder output. */

	int_fast16_t outbuf;
	/* The byte buffer (i.e., the B variable in the standard). */

	int_fast16_t lastbyte;
	/* The last byte output. */

	int err;
	/* The error indicator. */
	
} mqenc_t;

/* MQ arithmetic encoder state information. */
typedef struct {

	unsigned areg;
	/* The A register. */

	unsigned creg;
	/* The C register. */

	unsigned ctreg;
	/* The CT register. */

	int lastbyte;
	/* The last byte output by the encoder. */

} mqencstate_t;

/******************************************************************************\
* Functions and macros.
\******************************************************************************/

mqenc_t *mqenc_create(int maxctxs, stream_t *out);
/* Create an encoder. */

void mqenc_destroy(mqenc_t *enc);
/* Destroy an encoder. */

void mqenc_init(mqenc_t *enc);
/* Initialize an encoder. */

#define	mqenc_error(enc) \
	((enc)->err)

void mqenc_setctx(mqenc_t *enc, int ctxno, mqctx_t *ctx);
void mqenc_setctxs(mqenc_t *enc, int numctxs, mqctx_t *ctxs);

int mqenc_flush(mqenc_t *enc, int termmode);

void mqenc_getstate(mqenc_t *enc, mqencstate_t *state);
/* Get the current encoder state. */

#define	mqenc_setcurctx(enc, ctxno) \
        ((enc)->curctx = &(enc)->ctxs[ctxno]);
/* Set the current context. */

#if !defined(DEBUG)
#define	mqenc_putbit(enc, bit)	mqenc_putbit_macro(enc, bit)
#else
#define	mqenc_putbit(enc, bit)	mqenc_putbit_func(enc, bit)
#endif
/* Encode a bit. */

int mqenc_dump(mqenc_t *mqenc, FILE *out);

/******************************************************************************\
* Implementation-specific details.
\******************************************************************************/

/* Note: This macro is included only to satisfy the needs of
  the mqenc_putbit macro. */
#define	mqenc_putbit_macro(enc, bit) \
	(((*((enc)->curctx))->mps == (bit)) ? \
	  (((enc)->areg -= (*(enc)->curctx)->qeval), \
	  ((!((enc)->areg & 0x8000)) ? (mqenc_codemps2(enc)) : \
	  ((enc)->creg += (*(enc)->curctx)->qeval))) : \
	  mqenc_codelps(enc))

/* Note: These function prototypes are included only to satisfy the
  needs of the mqenc_putbit_macro macro.  Do not call any of these
  functions directly, or else... */
int mqenc_codemps2(mqenc_t *enc);
int mqenc_codelps(mqenc_t *enc);

/* Note: This function prototype is included only to satisfy the needs of
  the mqenc_putbit macro. */
int mqenc_putbit_func(mqenc_t *enc, int bit);

#endif
