/*
 * Copyright (c) 1999-2000 Image Power, Inc. and the University of
 *   British Columbia.
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * $Id$
 */

#ifndef T1COD_H
#define T1COD_H

/******************************************************************************\
* Includes.
\******************************************************************************/

#include "mqcod.h"
#include "stream.h"
#include "image.h"
#include "fix.h"
#include "tsfb.h"

/******************************************************************************\
* Constants.
\******************************************************************************/

#define NMSEDEC_BITS	7
#define NMSEDEC_FRACBITS	(NMSEDEC_BITS - 1)

#define SEG_INVALID	0
#define SEG_MQ		1
#define SEG_RAW		2

#define	PREC	32

#define	SIGPASS	0
#define	REFPASS	1
#define	CLNPASS	2

#define	NESIG	0x0001
#define	SESIG	0x0002
#define	SWSIG	0x0004
#define	NWSIG	0x0008
#define	NSIG	0x0010
#define	ESIG	0x0020
#define	SSIG	0x0040
#define	WSIG	0x0080
#define	NSGN	0x0100
#define	ESGN	0x0200
#define	SSGN	0x0400
#define	WSGN	0x0800

#define SIG	0x1000 /* sigma */
#define	REFINE	0x2000 /* sigma_cup */
#define	VISIT	0x4000 /* eta */

#define	OTHSIGMSK \
	(NSIG | NESIG | ESIG | SESIG | SSIG | SWSIG | WSIG | NWSIG)
#define	SGNMSK	(NSGN | ESGN | SSGN | WSGN)
#define	PRIMSIGMSK	(NSIG | ESIG | SSIG | WSIG)

#define	NUMAGGCTXS	1
#define	NUMZCCTXS	9
#define	NUMMAGCTXS	3
#define	NUMSCCTXS	5
#define	NUMUCTXS	1

#define	AGGCTXNO	0
#define	ZCCTXNO		(AGGCTXNO + NUMAGGCTXS)
#define	MAGCTXNO	(ZCCTXNO + NUMZCCTXS)
#define	SCCTXNO		(MAGCTXNO + NUMMAGCTXS)
#define	UCTXNO		(SCCTXNO + NUMSCCTXS)
#define	NUMCTXS		(UCTXNO + NUMUCTXS)

/******************************************************************************\
* Types.
\******************************************************************************/

/******************************************************************************\
* External data.
\******************************************************************************/

extern int zcctxnolut[];
extern int spblut[];
extern int scctxnolut[];
extern int magctxnolut[];
extern fix_t refnmsedec[];
extern fix_t signmsedec[];
extern fix_t refnmsedec0[];
extern fix_t signmsedec0[];

/******************************************************************************\
* Functions and macros.
\******************************************************************************/

void initctxs(mqctx_t *ctxs);
int getzcctxno(int f, int orient);
int getspb(int f);
int getscctxno(int f);
int getmagctxno(int f);

#define	GETZCCTXNO(f, orient) \
	(zcctxnolut[((orient) << 8) | ((f) & OTHSIGMSK)])
#define	GETSPB(f) \
	(spblut[((f) & (PRIMSIGMSK | SGNMSK)) >> 4])
#define	GETSCCTXNO(f) \
	(scctxnolut[((f) & (PRIMSIGMSK | SGNMSK)) >> 4])
#define	GETMAGCTXNO(f) \
	(magctxnolut[((f) & OTHSIGMSK) | ((((f) & REFINE) != 0) << 11)])

#define	NOMRANGEBITS(fmt, prec) \
	(((fmt) == FMT_UNSIGNED) ? ((prec) - 1) : (prec))

#define	ASR(x, n) \
	(((n) >= 0) ? ((x) >> (n)) : ((x) << (-(n))))
#define	ONES(n) \
	(((n) < 32) ? ((1 << (n)) - 1) : (0xffffffff))

#define	GETSIGNMSEDEC(x, bitpos)	getsignmsedec_macro(x, bitpos)
#define	GETREFNMSEDEC(x, bitpos)	getrefnmsedec_macro(x, bitpos)

fix_t getsignmsedec_func(fix_t x, int bitpos);
fix_t refsignmsedec_func(fix_t x, int bitpos);

#define	getsignmsedec_macro(x, bitpos) \
	((bitpos > NMSEDEC_FRACBITS) ? signmsedec[ASR(x, bitpos - NMSEDEC_FRACBITS) & ONES(NMSEDEC_BITS)] : \
	  (signmsedec0[ASR(x, bitpos - NMSEDEC_FRACBITS) & ONES(NMSEDEC_BITS)]))

#define	getrefnmsedec_macro(x, bitpos) \
	((bitpos > NMSEDEC_FRACBITS) ? refnmsedec[ASR(x, bitpos - NMSEDEC_FRACBITS) & ONES(NMSEDEC_BITS)] : \
	  (refnmsedec0[ASR(x, bitpos - NMSEDEC_FRACBITS) & ONES(NMSEDEC_BITS)]))

#define	UPDATEFLAGS4(fp, rowstep, s, vcausalflag) \
{ \
	register fix_t *np = (fp) - (rowstep); \
	register fix_t *sp = (fp) + (rowstep); \
	if ((vcausalflag)) { \
		sp[-1] |= NESIG; \
		sp[1] |= NWSIG; \
		if (s) { \
			*sp |= NSIG | NSGN; \
			(fp)[-1] |= ESIG | ESGN; \
			(fp)[1] |= WSIG | WSGN; \
		} else { \
			*sp |= NSIG; \
			(fp)[-1] |= ESIG; \
			(fp)[1] |= WSIG; \
		} \
	} else { \
		np[-1] |= SESIG; \
		np[1] |= SWSIG; \
		sp[-1] |= NESIG; \
		sp[1] |= NWSIG; \
		if (s) { \
			*np |= SSIG | SSGN; \
			*sp |= NSIG | NSGN; \
			(fp)[-1] |= ESIG | ESGN; \
			(fp)[1] |= WSIG | WSGN; \
		} else { \
			*np |= SSIG; \
			*sp |= NSIG; \
			(fp)[-1] |= ESIG; \
			(fp)[1] |= WSIG; \
		} \
	} \
}

void initluts();

int NOMINALGAIN(int qmfbid, int numlvls, int lvlno, int orient);
int PASSTYPE(int passno);
int SEGTYPE(int passno, int firstpassno, int bypass);
int SEGPASSCNT(int passno, int firstpassno, int numpasses, int bypass,
  int termall);
int ISTERMINATED(int passno, int firstpassno, int numpasses, int termall,
  int lazy);

double pow2i(int n);

extern mqctx_t mqctxs[];

#endif
