/*
 * Copyright (c) 1999-2000 Image Power, Inc. and the University of
 *   British Columbia.
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * Tier 1 Encoder
 *
 * $Id$
 */

/******************************************************************************\
* Includes.
\******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "t1enc.h"
#include "t1cod.h"
#include "jpcenc.h"
#include "fix.h"
#include "jpccod.h"
#include "memalloc.h"
#include "mathx.h"

/******************************************************************************\
* Code for encoding code blocks.
\******************************************************************************/

/* Encode all of the code blocks associated with the current tile. */
int enc_enccblks(enc_t *enc)
{
	enc_tcmpt_t *tcmpt;
	enc_tcmpt_t *endcomps;
	enc_rlvl_t *lvl;
	enc_rlvl_t *endlvls;
	enc_band_t *band;
	enc_band_t *endbands;
	enc_cblk_t *cblk;
	enc_cblk_t *endcblks;
	int i;
	int j;
	int mx;
	int bmx;
	int v;
	enc_tile_t *tile;
	uint_fast32_t prcno;
	enc_prc_t *prc;

	tile = enc->curtile;

	endcomps = &tile->tcmpts[tile->numtcmpts];
	for (tcmpt = tile->tcmpts; tcmpt != endcomps; ++tcmpt) {
		endlvls = &tcmpt->rlvls[tcmpt->numrlvls];
		for (lvl = tcmpt->rlvls; lvl != endlvls; ++lvl) {
			if (!lvl->bands) {
				continue;
			}
			endbands = &lvl->bands[lvl->numbands];
			for (band = lvl->bands; band != endbands; ++band) {
				if (!band->data) {
					continue;
				}
				for (prcno = 0, prc = band->prcs; prcno < lvl->numprcs; ++prcno, ++prc) {
					if (!prc->cblks) {
						continue;
					}
					bmx = 0;
					endcblks = &prc->cblks[prc->numcblks];
					for (cblk = prc->cblks; cblk != endcblks; ++cblk) {
						mx = 0;
						for (i = 0; i < matrix_numrows(cblk->data); ++i) {
							for (j = 0; j < matrix_numcols(cblk->data); ++j) {
								v = abs(matrix_get(cblk->data, i, j));
								if (v > mx) {
									mx = v;
								}
							}
						}
						if (mx > bmx) {
							bmx = mx;
						}
						cblk->numbps = max(firstone(mx) + 1 - NUMEXTRABITS, 0);
					}

					for (cblk = prc->cblks; cblk != endcblks; ++cblk) {
						cblk->numimsbs = band->numbps - cblk->numbps;
						assert(cblk->numimsbs >= 0);
					}

					for (cblk = prc->cblks; cblk != endcblks; ++cblk) {
						if (enc_enccblk(enc, cblk->stream, tcmpt, band, cblk)) {
							return -1;
						}
					}
				}
			}
		}
	}
	return 0;
}

int getthebyte(stream_t *in, long off)
{
	int c;
	long oldpos;
	oldpos = stream_tell(in);
	assert(oldpos >= 0);
	stream_seek(in, off, SEEK_SET);
	c = stream_peekc(in);
	stream_seek(in, oldpos, SEEK_SET);
	return c;
}

/* Encode a single code block. */
int enc_enccblk(enc_t *enc, stream_t *out, enc_tcmpt_t *tcmpt, enc_band_t *band, enc_cblk_t *cblk)
{
	enc_pass_t *pass;
	enc_pass_t *endpasses;
	int bitpos;
	int n;
	int adjust;
	int ret;
	int passtype;
	int t;
	bitstream_t *bout;
	enc_pass_t *termpass;
	enc_rlvl_t *rlvl;
	int vcausal;
	int segsym;
	int termmode;
	int c;

	bout = 0;
	rlvl = band->rlvl;

	cblk->stream = stream_memopen(0, 0);
	assert(cblk->stream);
	cblk->mqenc = mqenc_create(NUMCTXS, cblk->stream);
	assert(cblk->mqenc);
	mqenc_setctxs(cblk->mqenc, NUMCTXS, mqctxs);

	cblk->numpasses = (cblk->numbps > 0) ? (3 * cblk->numbps - 2) : 0;
	if (cblk->numpasses > 0) {
		cblk->passes = malloc(cblk->numpasses * sizeof(enc_pass_t));
		assert(cblk->passes);
	} else {
		cblk->passes = 0;
	}
	endpasses = &cblk->passes[cblk->numpasses];
	for (pass = cblk->passes; pass != endpasses; ++pass) {
		pass->start = 0;
		pass->end = 0;
		pass->term = ISTERMINATED(pass - cblk->passes, 0, cblk->numpasses, (tcmpt->cblksty & JPC_COX_TERMALL) != 0, (tcmpt->cblksty & JPC_COX_LAZY) != 0);
		pass->type = SEGTYPE(pass - cblk->passes, 0, (tcmpt->cblksty & JPC_COX_LAZY) != 0);
		pass->lyrno = -1;
if (pass == endpasses - 1) {
assert(pass->term == 1);
	pass->term = 1;
}
	}

	cblk->flags = matrix_create(matrix_numrows(cblk->data) + 2,
	  matrix_numcols(cblk->data) + 2);
	assert(cblk->flags);


	bitpos = cblk->numbps - 1;
	pass = cblk->passes;
	n = cblk->numpasses;
	while (--n >= 0) {

		if (pass->type == SEG_MQ) {
			/* NOP */
		} else {
			assert(pass->type == SEG_RAW);
			if (!bout) {
				bout = bitstream_sopen(cblk->stream, "w");
				assert(bout);
			}
		}

#if 1
		passtype = (pass - cblk->passes + 2) % 3;
#else
		passtype = PASSTYPE(pass - cblk->passes + 2);
#endif
		pass->start = stream_tell(cblk->stream);
#if 0
assert(stream_tell(cblk->stream) == stream_getrwcount(cblk->stream));
#endif
		assert(bitpos >= 0);
		vcausal = (tcmpt->cblksty & JPC_COX_VSC) != 0;
		segsym = (tcmpt->cblksty & JPC_COX_SEGSYM) != 0;
		if (pass->term) {
			termmode = ((tcmpt->cblksty & JPC_COX_PTERM) ?
			  MQENC_PTERM : MQENC_DEFTERM) + 1;
		} else {
			termmode = 0;
		}
		switch (passtype) {
		case SIGPASS:
			ret = (pass->type == SEG_MQ) ? enc_sigpass(cblk->mqenc,
			  bitpos, band->orient, vcausal, cblk->flags,
			  cblk->data, termmode, &pass->nmsedec) :
			  enc_rawsigpass(bout, bitpos, vcausal, cblk->flags,
			  cblk->data, termmode, &pass->nmsedec);
			break;
		case REFPASS:
			ret = (pass->type == SEG_MQ) ? enc_refpass(cblk->mqenc,
			  bitpos, vcausal, cblk->flags, cblk->data, termmode,
			  &pass->nmsedec) : enc_rawrefpass(bout, bitpos,
			  vcausal, cblk->flags, cblk->data, termmode,
			  &pass->nmsedec);
			break;
		case CLNPASS:
			assert(pass->type == SEG_MQ);
			ret = enc_clnpass(cblk->mqenc, bitpos, band->orient,
			  vcausal, segsym, cblk->flags, cblk->data, termmode,
			  &pass->nmsedec);
			break;
		default:
			assert(0);
			break;
		}

		if (pass->type == SEG_MQ) {
			if (pass->term) {
				mqenc_init(cblk->mqenc);
			}
			mqenc_getstate(cblk->mqenc, &pass->mqencstate);
			pass->end = stream_tell(cblk->stream);
			if (tcmpt->cblksty & JPC_COX_RESET) {
				mqenc_setctxs(cblk->mqenc, NUMCTXS, mqctxs);
			}
		} else {
			if (pass->term) {
				if (bitstream_pending(bout)) {
					bitstream_outalign(bout, 0x2a);
				}
				bout = 0;
				pass->end = stream_tell(cblk->stream);
			} else {
				pass->end = stream_tell(cblk->stream) +
				  bitstream_pending(bout);
/* NOTE - This will not work.  need to adjust by # of pending output bytes */
			}
		}
#if 0
/* XXX - This assertion fails sometimes when various coding modes are used.
This seems to be harmless, but why does it happen at all? */
assert(stream_tell(cblk->stream) == stream_getrwcount(cblk->stream));
#endif

		pass->wmsedec = fixtodbl(band->rlvl->tcmpt->synweight) *
		  fixtodbl(band->rlvl->tcmpt->synweight) *
		  fixtodbl(band->synweight) *
		  fixtodbl(band->synweight) *
		  fixtodbl(band->absstepsize) * fixtodbl(band->absstepsize) *
		  ((double) (1 << bitpos)) * ((double)(1 << bitpos)) *
		  fixtodbl(pass->nmsedec);
		pass->cumwmsedec = pass->wmsedec;
		if (pass != cblk->passes) {
			pass->cumwmsedec += pass[-1].cumwmsedec;
		}
		if (passtype == CLNPASS) {
			--bitpos;
		}
		++pass;
	}

#if 0
dump_passes(cblk->passes, cblk->numpasses, cblk);
#endif

	n = 0;
	endpasses = &cblk->passes[cblk->numpasses];
	for (pass = cblk->passes; pass != endpasses; ++pass) {
		if (pass->start < n) {
			pass->start = n;
		}
		if (pass->end < n) {
			pass->end = n;
		}
		if (!pass->term) {
			termpass = pass;
			while (termpass - pass < cblk->numpasses &&
			  !termpass->term) {
				++termpass;
			}
			if (pass->type == SEG_MQ) {
				t = (pass->mqencstate.lastbyte == 0xff) ? 1 : 0;
				if (pass->mqencstate.ctreg >= 5) {
					adjust = 4 + t;
				} else {
					adjust = 5 + t;
				}
				pass->end += adjust;
			}
			if (pass->end > termpass->end) {
				pass->end = termpass->end;
			}
			if ((c = getthebyte(cblk->stream, pass->end - 1)) == EOF) {
				abort();
			}
			if (c == 0xff) {
				++pass->end;
			}
			n = max(n, pass->end);
		} else {
			n = max(n, pass->end);
		}
	}

#if 0
dump_passes(cblk->passes, cblk->numpasses, cblk);
#endif

	return 0;
}

/******************************************************************************\
* Code for significance pass.
\******************************************************************************/

#define	sigpass_step(fp, frowstep, dp, bitpos, one, nmsedec, orient, mqenc, vcausalflag) \
{ \
	int f; \
	int v; \
	f = *(fp); \
	if ((f & OTHSIGMSK) && !(f & (SIG | VISIT))) { \
		v = (abs(*(dp)) & (one)) ? 1 : 0; \
		mqenc_setcurctx(mqenc, GETZCCTXNO(f, (orient))); \
		mqenc_putbit(mqenc, v); \
		if (v) { \
			*(nmsedec) += GETSIGNMSEDEC(abs(*(dp)), (bitpos) + NUMEXTRABITS); \
			v = ((*(dp) < 0) ? 1 : 0); \
			mqenc_setcurctx(mqenc, GETSCCTXNO(f)); \
			mqenc_putbit(mqenc, v ^ GETSPB(f)); \
			UPDATEFLAGS4(fp, frowstep, v, vcausalflag); \
			*(fp) |= SIG; \
		} \
		*(fp) |= VISIT; \
	} \
}

int enc_sigpass(mqenc_t *mqenc, int bitpos, int orient, int vcausalflag,
  matrix_t *flags, matrix_t *data, int term, long *nmsedec)
{
	int i;
	int j;
	int one;
	int vscanlen;
	int width;
	int height;
	int frowstep;
	int drowstep;
	int fstripestep;
	int dstripestep;
	fix_t *fstripestart;
	fix_t *dstripestart;
	fix_t *fp;
	fix_t *dp;
	fix_t *fvscanstart;
	fix_t *dvscanstart;
	int k;

	*nmsedec = 0;
	width = matrix_numcols(data);
	height = matrix_numrows(data);
	frowstep = matrix_rowstep(flags);
	drowstep = matrix_rowstep(data);
	fstripestep = frowstep << 2;
	dstripestep = drowstep << 2;

	one = 1 << (bitpos + NUMEXTRABITS);

	fstripestart = matrix_getref(flags, 1, 1);
	dstripestart = matrix_getref(data, 0, 0);
	for (i = height; i > 0; i -= 4, fstripestart += fstripestep,
	  dstripestart += dstripestep) {
		fvscanstart = fstripestart;
		dvscanstart = dstripestart;
		vscanlen = min(i, 4);
		for (j = width; j > 0; --j, ++fvscanstart, ++dvscanstart) {
			fp = fvscanstart;
			dp = dvscanstart;
			k = vscanlen;

			sigpass_step(fp, frowstep, dp, bitpos, one,
			  nmsedec, orient, mqenc, vcausalflag);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;
			sigpass_step(fp, frowstep, dp, bitpos, one,
			  nmsedec, orient, mqenc, 0);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;
			sigpass_step(fp, frowstep, dp, bitpos, one,
			  nmsedec, orient, mqenc, 0);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;
			sigpass_step(fp, frowstep, dp, bitpos, one,
			  nmsedec, orient, mqenc, 0);

		}
	}

	if (term) {
		mqenc_flush(mqenc, term - 1);
	}

	return mqenc_error(mqenc) ? (-1) : 0;
}

#define	rawsigpass_step(fp, frowstep, dp, bitpos, one, nmsedec, out, vcausalflag) \
{ \
	fix_t f = *(fp); \
	fix_t v; \
	if ((f & OTHSIGMSK) && !(f & (SIG | VISIT))) { \
		v = (abs(*(dp)) & (one)) ? 1 : 0; \
		if ((bitstream_putbit((out), v)) == EOF) { \
			return -1; \
		} \
		if (v) { \
			*(nmsedec) += GETSIGNMSEDEC(abs(*(dp)), (bitpos) + NUMEXTRABITS); \
			v = ((*(dp) < 0) ? 1 : 0); \
			if (bitstream_putbit(out, v) == EOF) { \
				return -1; \
			} \
			UPDATEFLAGS4(fp, frowstep, v, vcausalflag); \
			*(fp) |= SIG; \
		} \
		*(fp) |= VISIT; \
	} \
}

int enc_rawsigpass(bitstream_t *out, int bitpos, int vcausalflag, matrix_t *flags,
  matrix_t *data, int term, long *nmsedec)
{
	int i;
	int j;
	int k;
	int one;
	int vscanlen;
	int width;
	int height;
	int frowstep;
	int drowstep;
	int fstripestep;
	int dstripestep;
	fix_t *fstripestart;
	fix_t *dstripestart;
	fix_t *fp;
	fix_t *dp;
	fix_t *fvscanstart;
	fix_t *dvscanstart;

	*nmsedec = 0;
	width = matrix_numcols(data);
	height = matrix_numrows(data);
	frowstep = matrix_rowstep(flags);
	drowstep = matrix_rowstep(data);
	fstripestep = frowstep << 2;
	dstripestep = drowstep << 2;

	one = 1 << (bitpos + NUMEXTRABITS);

	fstripestart = matrix_getref(flags, 1, 1);
	dstripestart = matrix_getref(data, 0, 0);
	for (i = height; i > 0; i -= 4, fstripestart += fstripestep,
	  dstripestart += dstripestep) {
		fvscanstart = fstripestart;
		dvscanstart = dstripestart;
		vscanlen = min(i, 4);
		for (j = width; j > 0; --j, ++fvscanstart, ++dvscanstart) {
			fp = fvscanstart;
			dp = dvscanstart;
			k = vscanlen;

			rawsigpass_step(fp, frowstep, dp, bitpos, one,
			  nmsedec, out, vcausalflag);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;

			rawsigpass_step(fp, frowstep, dp, bitpos, one,
			  nmsedec, out, 0);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;

			rawsigpass_step(fp, frowstep, dp, bitpos, one,
			  nmsedec, out, 0);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;

			rawsigpass_step(fp, frowstep, dp, bitpos, one,
			  nmsedec, out, 0);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;

		}
	}

	if (term) {
		bitstream_outalign(out, 0x2a);
	}

	return 0;
}

/******************************************************************************\
* Code for refinement pass.
\******************************************************************************/

#define	refpass_step(fp, dp, bitpos, one, nmsedec, mqenc, vcausalflag) \
{ \
	int v; \
	if (((*(fp)) & (SIG | VISIT)) == SIG) { \
		(d) = *(dp); \
		*(nmsedec) += GETREFNMSEDEC(abs(d), (bitpos) + NUMEXTRABITS); \
		mqenc_setcurctx((mqenc), GETMAGCTXNO(*(fp))); \
		v = (abs(d) & (one)) ? 1 : 0; \
		mqenc_putbit((mqenc), v); \
		*(fp) |= REFINE; \
	} \
}

int enc_refpass(mqenc_t *mqenc, int bitpos, int vcausalflag, matrix_t *flags, matrix_t *data,
  int term, long *nmsedec)
{
	int i;
	int j;
	int one;
	int vscanlen;
	int d;
	int width;
	int height;
	int frowstep;
	int drowstep;
	int fstripestep;
	int dstripestep;
	fix_t *fstripestart;
	fix_t *dstripestart;
	fix_t *fvscanstart;
	fix_t *dvscanstart;
	fix_t *dp;
	fix_t *fp;
int k;

	*nmsedec = 0;
	width = matrix_numcols(data);
	height = matrix_numrows(data);
	frowstep = matrix_rowstep(flags);
	drowstep = matrix_rowstep(data);
	fstripestep = frowstep << 2;
	dstripestep = drowstep << 2;

	one = 1 << (bitpos + NUMEXTRABITS);

	fstripestart = matrix_getref(flags, 1, 1);
	dstripestart = matrix_getref(data, 0, 0);
	for (i = height; i > 0; i -= 4, fstripestart += fstripestep,
	  dstripestart += dstripestep) {
		fvscanstart = fstripestart;
		dvscanstart = dstripestart;
		vscanlen = min(i, 4);
		for (j = width; j > 0; --j, ++fvscanstart, ++dvscanstart) {
			fp = fvscanstart;
			dp = dvscanstart;
			k = vscanlen;

			refpass_step(fp, dp, bitpos, one, nmsedec,
			  mqenc, vcausalflag);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;
			refpass_step(fp, dp, bitpos, one, nmsedec,
			  mqenc, 0);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;
			refpass_step(fp, dp, bitpos, one, nmsedec,
			  mqenc, 0);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;
			refpass_step(fp, dp, bitpos, one, nmsedec,
			  mqenc, 0);

		}
	}

	if (term) {
		mqenc_flush(mqenc, term - 1);
	}

	return mqenc_error(mqenc) ? (-1) : 0;
}

#define	rawrefpass_step(fp, dp, bitpos, one, nmsedec, out, vcausalflag) \
{ \
	fix_t d; \
	fix_t v; \
	if (((*(fp)) & (SIG | VISIT)) == SIG) { \
		d = *(dp); \
		*(nmsedec) += GETREFNMSEDEC(abs(d), (bitpos) + NUMEXTRABITS); \
		v = (abs(d) & (one)) ? 1 : 0; \
		if (bitstream_putbit((out), v) == EOF) { \
			return -1; \
		} \
		*(fp) |= REFINE; \
	} \
}

int enc_rawrefpass(bitstream_t *out, int bitpos, int vcausalflag, matrix_t *flags,
  matrix_t *data, int term, long *nmsedec)
{
	int i;
	int j;
	int k;
	int one;
	int vscanlen;
	int width;
	int height;
	int frowstep;
	int drowstep;
	int fstripestep;
	int dstripestep;
	fix_t *fstripestart;
	fix_t *dstripestart;
	fix_t *fvscanstart;
	fix_t *dvscanstart;
	fix_t *dp;
	fix_t *fp;

	*nmsedec = 0;
	width = matrix_numcols(data);
	height = matrix_numrows(data);
	frowstep = matrix_rowstep(flags);
	drowstep = matrix_rowstep(data);
	fstripestep = frowstep << 2;
	dstripestep = drowstep << 2;

	one = 1 << (bitpos + NUMEXTRABITS);

	fstripestart = matrix_getref(flags, 1, 1);
	dstripestart = matrix_getref(data, 0, 0);
	for (i = height; i > 0; i -= 4, fstripestart += fstripestep,
	  dstripestart += dstripestep) {
		fvscanstart = fstripestart;
		dvscanstart = dstripestart;
		vscanlen = min(i, 4);
		for (j = width; j > 0; --j, ++fvscanstart, ++dvscanstart) {
			fp = fvscanstart;
			dp = dvscanstart;
			k = vscanlen;

			rawrefpass_step(fp, dp, bitpos, one, nmsedec,
			  out, vcausalflag);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;
			rawrefpass_step(fp, dp, bitpos, one, nmsedec,
			  out, vcausalflag);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;
			rawrefpass_step(fp, dp, bitpos, one, nmsedec,
			  out, vcausalflag);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;
			rawrefpass_step(fp, dp, bitpos, one, nmsedec,
			  out, vcausalflag);

		}
	}

	if (term) {
		bitstream_outalign(out, 0x2a);
	}

	return 0;
}

/******************************************************************************\
* Code for cleanup pass.
\******************************************************************************/

#define	clnpass_step(fp, frowstep, dp, bitpos, one, orient, nmsedec, mqenc, label1, label2, vcausalflag) \
{ \
	int f; \
	int v; \
label1 \
	f = *(fp); \
	if (!(f & (SIG | VISIT))) { \
		mqenc_setcurctx(mqenc, GETZCCTXNO(f, (orient))); \
		v = (abs(*(dp)) & (one)) ? 1 : 0; \
		mqenc_putbit((mqenc), v); \
		if (v) { \
label2 \
			f = *(fp); \
			/* Coefficient is significant. */ \
			*(nmsedec) += GETSIGNMSEDEC(abs(*(dp)), (bitpos) + NUMEXTRABITS); \
			mqenc_setcurctx((mqenc), GETSCCTXNO(f)); \
			v = ((*(dp) < 0) ? 1 : 0); \
			mqenc_putbit((mqenc), v ^ GETSPB(f)); \
			UPDATEFLAGS4((fp), (frowstep), v, vcausalflag); \
			*(fp) |= SIG; \
		} \
	} \
	*(fp) &= ~VISIT; \
}

int enc_clnpass(mqenc_t *mqenc, int bitpos, int orient, int vcausalflag, int segsymflag, matrix_t *flags,
  matrix_t *data, int term, long *nmsedec)
{
	int i;
	int j;
	int k;
	int vscanlen;
	int v;
	int runlen;
	fix_t *fp;
	int width;
	int height;
	fix_t *dp;
	int one;
	int frowstep;
	int drowstep;
	int fstripestep;
	int dstripestep;
	fix_t *fstripestart;
	fix_t *dstripestart;
	fix_t *fvscanstart;
	fix_t *dvscanstart;

	*nmsedec = 0;
	width = matrix_numcols(data);
	height = matrix_numrows(data);
	frowstep = matrix_rowstep(flags);
	drowstep = matrix_rowstep(data);
	fstripestep = frowstep << 2;
	dstripestep = drowstep << 2;

	one = 1 << (bitpos + NUMEXTRABITS);

	fstripestart = matrix_getref(flags, 1, 1);
	dstripestart = matrix_getref(data, 0, 0);
	for (i = height; i > 0; i -= 4, fstripestart += fstripestep,
	  dstripestart += dstripestep) {
		fvscanstart = fstripestart;
		dvscanstart = dstripestart;
		vscanlen = min(i, 4);
		for (j = width; j > 0; --j, ++fvscanstart, ++dvscanstart) {

			fp = fvscanstart;
			if (vscanlen >= 4 && !((*fp) & (SIG | VISIT |
			  OTHSIGMSK)) && (fp += frowstep, !((*fp) & (SIG |
			  VISIT | OTHSIGMSK))) && (fp += frowstep, !((*fp) &
			  (SIG | VISIT | OTHSIGMSK))) && (fp += frowstep,
			  !((*fp) & (SIG | VISIT | OTHSIGMSK)))) {
				dp = dvscanstart;
				for (k = 0; k < vscanlen; ++k) {
					v = (abs(*dp) & one) ? 1 : 0;
					if (v) {
						break;
					}
					dp += drowstep;
				}
				runlen = k;
				if (runlen >= 4) {
					mqenc_setcurctx(mqenc, AGGCTXNO);
					mqenc_putbit(mqenc, 0);
					continue;
				}
				mqenc_setcurctx(mqenc, AGGCTXNO);
				mqenc_putbit(mqenc, 1);
				mqenc_setcurctx(mqenc, UCTXNO);
				mqenc_putbit(mqenc, runlen >> 1);
				mqenc_putbit(mqenc, runlen & 1);
				fp = fvscanstart + frowstep * runlen;
				dp = dvscanstart + drowstep * runlen;
				k = vscanlen - runlen;
				switch (runlen) {
				case 0:
					goto clnpass_partial0;
					break;
				case 1:
					goto clnpass_partial1;
					break;
				case 2:
					goto clnpass_partial2;
					break;
				case 3:
					goto clnpass_partial3;
					break;
				}
			} else {
				runlen = 0;
				fp = fvscanstart;
				dp = dvscanstart;
				k = vscanlen;
				goto clnpass_full0;
			}
			clnpass_step(fp, frowstep, dp, bitpos, one,
			  orient, nmsedec, mqenc, clnpass_full0:, clnpass_partial0:, vcausalflag);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;
			clnpass_step(fp, frowstep, dp, bitpos, one,
				orient, nmsedec, mqenc, ;, clnpass_partial1:, 0);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;
			clnpass_step(fp, frowstep, dp, bitpos, one,
				orient, nmsedec, mqenc, ;, clnpass_partial2:, 0);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;
			clnpass_step(fp, frowstep, dp, bitpos, one,
				orient, nmsedec, mqenc, ;, clnpass_partial3:, 0);
		}
	}

	if (segsymflag) {
		mqenc_setcurctx(mqenc, UCTXNO);
		mqenc_putbit(mqenc, 1);
		mqenc_putbit(mqenc, 0);
		mqenc_putbit(mqenc, 1);
		mqenc_putbit(mqenc, 0);
	}

	if (term) {
		mqenc_flush(mqenc, term - 1);
	}

	return mqenc_error(mqenc) ? (-1) : 0;
}
