/*
 * Copyright (c) 1999-2000 Image Power, Inc. and the University of
 *   British Columbia.
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * Tier 2 Decoder
 *
 * $Id$
 */

/******************************************************************************\
* Includes.
\******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "types.h"
#include "fix.h"
#include "bs.h"
#include "stream.h"
#include "jpcdec.h"
#include "cs.h"
#include "mqdec.h"
#include "t2dec.h"
#include "memalloc.h"
#include "util.h"
#include "t1cod.h"
#include "mathx.h"

/******************************************************************************\
*
\******************************************************************************/

long dec_lookahead(stream_t *in);
static int getcommacode(bitstream_t *in);
static int getnumnewpasses(bitstream_t *in);
static int dec_decodepkt(jpd_t *dec, stream_t *pkthdrstream, stream_t *in, int compno, int lvlno,
  int prcno, int lyrno);

/******************************************************************************\
* Code.
\******************************************************************************/

static int getcommacode(bitstream_t *in)
{
	int n;
	int v;

	n = 0;
	for (;;) {
		if ((v = bitstream_getbit(in)) < 0) {
			return -1;
		}
		if (bitstream_eof(in)) {
			return -1;
		}
		if (!v) {
			break;
		}
		++n;
	}

	return n;
}

static int getnumnewpasses(bitstream_t *in)
{
	int n;

	if ((n = bitstream_getbit(in)) > 0) {
		if ((n = bitstream_getbit(in)) > 0) {
			if ((n = bitstream_getbits(in, 2)) == 3) {
				if ((n = bitstream_getbits(in, 5)) == 31) {
					if ((n = bitstream_getbits(in, 7)) >= 0) {
						n += 36 + 1;
					}
				} else if (n >= 0) {
					n += 5 + 1;
				}
			} else if (n >= 0) {
				n += 2 + 1;
			}
		} else if (!n) {
			n += 2;
		}
	} else if (!n) {
		++n;
	}

	return n;
}

static int dec_decodepkt(jpd_t *dec, stream_t *pkthdrstream, stream_t *in, int compno, int rlvlno,
  int prcno, int lyrno)
{
	bitstream_t *inb;
	jpd_tcomp_t *tcomp;
	jpd_rlvl_t *rlvl;
	jpd_band_t *band;
	jpd_cblk_t *cblk;
	int n;
	int m;
	int i;
	tagtreenode_t *leaf;
	int included;
	int ret;
	int numnewpasses;
	jpd_seg_t *seg;
	int len;
	int present;
	int savenumnewpasses;
	int mycounter;
	jpc_ms_t *ms;
	jpd_tile_t *tile;
	jpd_ccp_t *ccp;
	jpd_cp_t *cp;
	int bandno;
	jpd_prc_t *prc;
	int usedcblkcnt;
	int cblkno;
	uint_fast32_t bodylen;
	bool discard;
	int passno;
	int maxpasses;
	int hdrlen;
	int hdroffstart;
	int hdroffend;

	discard = (lyrno >= dec->maxlyrs);

	tile = dec->curtile;
	cp = tile->cp;
	ccp = &cp->ccps[compno];

	/*
	 * Decode the packet header.
	 */

	/* Decode the SOP marker segment if present. */
	if (cp->csty & JPC_COD_SOP) {
		if (dec_lookahead(in) == JPC_MS_SOP) {
			if (!(ms = jpc_getms(in, dec->cstate))) {
				return -1;
			}
			if (jpc_ms_gettype(ms) != JPC_MS_SOP) {
				jpc_ms_destroy(ms);
				fprintf(stderr, "missing SOP marker segment\n");
				return -1;
			}
			jpc_ms_destroy(ms);
		}
	}

hdroffstart = stream_getrwcount(pkthdrstream);

	if (!(inb = bitstream_sopen(pkthdrstream, "r"))) {
		return -1;
	}

	if ((present = bitstream_getbit(inb)) < 0) {
		return 1;
	}
	LOGMSG(10, ("\n", present));
	LOGMSG(10, ("present=%d ", present));

	/* Is the packet non-empty? */
	if (present) {
		/* The packet is non-empty. */
		tcomp = &tile->tcomps[compno];
		rlvl = &tcomp->rlvls[rlvlno];
		bodylen = 0;
		for (bandno = 0, band = rlvl->bands; bandno < rlvl->numbands;
		  ++bandno, ++band) {
			if (!band->data) {
				continue;
			}
			prc = &band->prcs[prcno];
			if (!prc->cblks) {
				continue;
			}
			usedcblkcnt = 0;
			for (cblkno = 0, cblk = prc->cblks; cblkno < prc->numcblks;
			  ++cblkno, ++cblk) {
				++usedcblkcnt;
				if (!cblk->numpasses) {
					leaf = tagtree_getleaf(prc->incltagtree, usedcblkcnt - 1);
					if ((included = tagtree_decode(prc->incltagtree, leaf, lyrno + 1, inb)) < 0) {
						return -1;
					}
				} else {
					if ((included = bitstream_getbit(inb)) < 0) {
						return -1;
					}
				}
				LOGMSG(10, ("\n"));
				LOGMSG(10, ("included=%d ", included));
				if (!included) {
					continue;
				}
				if (!cblk->numpasses) {
					i = 1;
					leaf = tagtree_getleaf(prc->numimsbstagtree, usedcblkcnt - 1);
					for (;;) {
						if ((ret = tagtree_decode(prc->numimsbstagtree, leaf, i, inb)) < 0) {
							return -1;
						}
						if (ret) {
							break;
						}
						++i;
					}
					cblk->numimsbs = i - 1;
					cblk->firstpassno = cblk->numimsbs * 3;
				}
				if ((numnewpasses = getnumnewpasses(inb)) < 0) {
					return -1;
				}
				LOGMSG(10, ("numnewpasses=%d ", numnewpasses));
				seg = cblk->curseg;
				savenumnewpasses = numnewpasses;
				mycounter = 0;
				if (numnewpasses > 0) {
					if ((m = getcommacode(inb)) < 0) {
						return -1;
					}
					cblk->numlenbits += m;
					LOGMSG(10, ("increment=%d ", m));
					while (numnewpasses > 0) {
						passno = cblk->firstpassno + cblk->numpasses + mycounter;
	/* XXX - the maxpasses is not set correctly */
						maxpasses = SEGPASSCNT(passno, cblk->firstpassno, 10000, (ccp->cblkctx & JPC_COX_LAZY) != 0, (ccp->cblkctx & JPC_COX_TERMALL) != 0);
						if (!discard && !seg) {
							if (!(seg = seg_alloc())) {
								return -1;
							}
							seglist_insert(&cblk->segs, cblk->segs.tail, seg);
							if (!cblk->curseg) {
								cblk->curseg = seg;
							}
							seg->passno = passno;
							seg->type = SEGTYPE(seg->passno, cblk->firstpassno, (ccp->cblkctx & JPC_COX_LAZY) != 0);
							seg->maxpasses = maxpasses;
						}
						n = min(numnewpasses, maxpasses);
						mycounter += n;
						numnewpasses -= n;
						if ((len = bitstream_getbits(inb, cblk->numlenbits + floorlog2(n))) < 0) {
							return -1;
						}
						LOGMSG(10, ("len=%d ", len));
						if (!discard) {
							seg->lyrno = lyrno;
							seg->numpasses += n;
							seg->cnt = len;
							seg = seg->next;
						}
						bodylen += len;
					}
				}
				cblk->numpasses += savenumnewpasses;
			}
		}

		bitstream_inalign(inb, 0, 0);
		bitstream_close(inb);

	} else {
		if (bitstream_inalign(inb, 0x7f, 0)) {
			fprintf(stderr, "alignment failed\n");
			return -1;
		}
	}

	hdroffend = stream_getrwcount(pkthdrstream);
	hdrlen = hdroffend - hdroffstart;
	if (getdbglevel() >= 5) {
		fprintf(stderr, "hdrlen=%lu bodylen=%lu \n", (unsigned long) hdrlen,
		  (unsigned long) bodylen);
	}

	if (cp->csty & JPC_COD_EPH) {
		if (dec_lookahead(pkthdrstream) == JPC_MS_EPH) {
			if (!(ms = jpc_getms(pkthdrstream, dec->cstate))) {
				fprintf(stderr, "cannot get (EPH) marker segment\n");
				return -1;
			}
			if (jpc_ms_gettype(ms) != JPC_MS_EPH) {
				jpc_ms_destroy(ms);
				fprintf(stderr, "missing EPH marker segment\n");
				return -1;
			}
			jpc_ms_destroy(ms);
		}
	}

	/* decode the packet body. */

	if (getdbglevel() >= 1) {
		fprintf(stderr, "packet body offset=%06ld\n", (long) stream_getrwcount(in));
	}

	if (!discard) {
		tcomp = &tile->tcomps[compno];
		rlvl = &tcomp->rlvls[rlvlno];
		for (bandno = 0, band = rlvl->bands; bandno < rlvl->numbands;
		  ++bandno, ++band) {
			if (!band->data) {
				continue;
			}
			prc = &band->prcs[prcno];
			if (!prc->cblks) {
				continue;
			}
			for (cblkno = 0, cblk = prc->cblks; cblkno < prc->numcblks;
			  ++cblkno, ++cblk) {
				seg = cblk->curseg;
				while (seg) {
					if (!seg->stream) {
						if (!(seg->stream = stream_memopen(0, 0))) {
							return -1;
						}
					}
#if 0
fprintf(stderr, "lyrno=%02d, compno=%02d, lvlno=%02d, prcno=%02d, bandno=%02d, cblkno=%02d, passno=%02d numpasses=%02d cnt=%d numbps=%d, numimsbs=%d\n", lyrno, compno, rlvlno, prcno, band - rlvl->bands, cblk - prc->cblks, seg->passno, seg->numpasses, seg->cnt, band->numbps, cblk->numimsbs);
#endif
					if (seg->cnt > 0) {
						if (jpc_getdata(in, seg->stream, seg->cnt) < 0) {
							return -1;
						}
						seg->cnt = 0;
					}
					if (seg->numpasses >= seg->maxpasses) {
						cblk->curseg = seg->next;
					}
					seg = seg->next;
				}
			}
		}
	} else {
		if (stream_gobble(in, bodylen) != bodylen) {
			return -1;
		}
	}
	return 0;
}

/********************************************************************************************/
/********************************************************************************************/

int dec_decodepkts(jpd_t *dec, stream_t *pkthdrstream, stream_t *in)
{
	jpd_tile_t *tile;
	pi_t *pi;
	int ret;

	tile = dec->curtile;
	pi = tile->pi;
	for (;;) {
if (!tile->pkthdrstream || stream_peekc(tile->pkthdrstream) == EOF) {
		switch (dec_lookahead(in)) {
		case JPC_MS_EOC:
		case JPC_MS_SOT:
			return 0;
			break;
		case JPC_MS_SOP:
		case JPC_MS_EPH:
		case 0:
			break;
		default:
			return -1;
			break;
		}
}
		if ((ret = pi_next(pi))) {
			return ret;
		}
		if (getdbglevel() >= 1) {
			fprintf(stderr, "packet offset=%08ld prg=%d cmptno=%02d "
			  "rlvlno=%02d prcno=%03d lyrno=%02d\n", (long)
			  stream_getrwcount(in), pi_prg(pi), pi_cmptno(pi),
			  pi_rlvlno(pi), pi_prcno(pi), pi_lyrno(pi));
		}
		if (dec_decodepkt(dec, pkthdrstream, in, pi_cmptno(pi), pi_rlvlno(pi),
		  pi_prcno(pi), pi_lyrno(pi))) {
			return -1;
		}
	}

	return 0;
}

pi_t *pi_create(jpd_t *dec, jpd_tile_t *tile)
{
	pi_t *pi;
	int compno;
	picomp_t *picomp;
	pirlvl_t *pirlvl;
	jpd_tcomp_t *tcomp;
	int rlvlno;
	jpd_rlvl_t *rlvl;
	int prcno;
	int *prclyrno;
	dec_cmpt_t *cmpt;

	if (!(pi = pi_create0())) {
		return 0;
	}
	pi->numcomps = dec->numcomps;
	if (!(pi->picomps = malloc(pi->numcomps * sizeof(picomp_t)))) {
		pi_destroy(pi);
		return 0;
	}
	for (compno = 0, picomp = pi->picomps; compno < pi->numcomps; ++compno,
	  ++picomp) {
		picomp->pirlvls = 0;
	}

	for (compno = 0, tcomp = tile->tcomps, picomp = pi->picomps;
	  compno < pi->numcomps; ++compno, ++tcomp, ++picomp) {
		picomp->numrlvls = tcomp->numrlvls;
		if (!(picomp->pirlvls = malloc(picomp->numrlvls *
		  sizeof(pirlvl_t)))) {
			pi_destroy(pi);
			return 0;
		}
		for (rlvlno = 0, pirlvl = picomp->pirlvls; rlvlno <
		  picomp->numrlvls; ++rlvlno, ++pirlvl) {
			pirlvl->prclyrnos = 0;
		}
		for (rlvlno = 0, pirlvl = picomp->pirlvls, rlvl = tcomp->rlvls;
		  rlvlno < picomp->numrlvls; ++rlvlno, ++pirlvl, ++rlvl) {
/* XXX sizeof(long) should be sizeof different type */
			pirlvl->numprcs = rlvl->numprcs;
			if (!(pirlvl->prclyrnos = malloc(pirlvl->numprcs *
			  sizeof(long)))) {
				pi_destroy(pi);
				return 0;
			}
		}
	}

	pi->maxrlvls = 0;
	for (compno = 0, tcomp = tile->tcomps, picomp = pi->picomps, cmpt =
	  dec->cmpts; compno < pi->numcomps; ++compno, ++tcomp, ++picomp,
	  ++cmpt) {
		picomp->hsamp = cmpt->hstep;
		picomp->vsamp = cmpt->vstep;
		for (rlvlno = 0, pirlvl = picomp->pirlvls, rlvl = tcomp->rlvls;
		  rlvlno < picomp->numrlvls; ++rlvlno, ++pirlvl, ++rlvl) {
			pirlvl->prcwidthexpn = rlvl->prcwidthexpn;
			pirlvl->prcheightexpn = rlvl->prcheightexpn;
			for (prcno = 0, prclyrno = pirlvl->prclyrnos;
			  prcno < pirlvl->numprcs; ++prcno, ++prclyrno) {
				*prclyrno = 0;
			}
			pirlvl->numhprcs = rlvl->numhprcs;
		}
		if (pi->maxrlvls < tcomp->numrlvls) {
			pi->maxrlvls = tcomp->numrlvls;
		}
	}

	pi->numlyrs = tile->cp->numlyrs;
	pi->xstart = tile->xstart;
	pi->ystart = tile->ystart;
	pi->xend = tile->xend;
	pi->yend = tile->yend;

	pi->picomp = 0;
	pi->pirlvl = 0;
	pi->x = 0;
	pi->y = 0;
	pi->compno = 0;
	pi->rlvlno = 0;
	pi->prcno = 0;
	pi->lyrno = 0;
	pi->xstep = 0;
	pi->ystep = 0;

	pi->pchgno = -1;

	pi->defaultpchg.prgord = tile->cp->prgord;
	pi->defaultpchg.compnostart = 0;
	pi->defaultpchg.compnoend = pi->numcomps;
	pi->defaultpchg.rlvlnostart = 0;
	pi->defaultpchg.rlvlnoend = pi->maxrlvls;
	pi->defaultpchg.lyrnoend = pi->numlyrs;
	pi->pchg = 0;

	pi->valid = 0;

	return pi;
}

long dec_lookahead(stream_t *in)
{
	uint_fast16_t x;
	if (jpc_getuint16(in, &x)) {
		return -1;
	}
	if (stream_ungetc(in, x & 0xff) == EOF ||
	  stream_ungetc(in, x >> 8) == EOF) {
		return -1;
	}
	if (x >= JPC_MS_INMIN && x <= JPC_MS_INMAX) {
		return x;
	}
	return 0;
}
