/*
 * Copyright (c) 1999-2000 Image Power, Inc. and the University of
 *   British Columbia.
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * Tag Tree Library
 *
 * $Id$
 */

#ifndef TAGTREE_H
#define TAGTREE_H

/******************************************************************************\
* Includes
\******************************************************************************/

#include <limits.h>
#include <stdio.h>
#include "bs.h"

/******************************************************************************\
* Constants
\******************************************************************************/

#define TAGTREE_MAXDEPTH	32
/* The maximum allowable depth for a tag tree. */

/******************************************************************************\
* Types
\******************************************************************************/

/*
 * Tag tree node.
 */

typedef struct tagtreenode_ {

	struct tagtreenode_ *parent_;
	/* The parent of this node. */

	int value_;
	/* The value associated with this node. */

	int low_;
	/* The lower bound on the value associated with this node. */

	int known_;
	/* A flag indicating if the value is known exactly. */

} tagtreenode_t;

/*
 * Tag tree.
 */

typedef struct {

	int numleafsh_;
	/* The number of leaves in the horizontal direction. */

	int numleafsv_;
	/* The number of leaves in the vertical direction. */

	int numnodes_;
	/* The total number of nodes in the tree. */

	tagtreenode_t *nodes_;
	/* The nodes. */

} tagtree_t;

/******************************************************************************\
* Functions.
\******************************************************************************/

tagtree_t *tagtree_create(int numleafsh, int numleafsv);
/* Create a tag tree. */

void tagtree_destroy(tagtree_t *tree);
/* Destroy a tag tree. */

void tagtree_copy(tagtree_t *dsttree, tagtree_t *srctree);
/* Copy data from one tag tree to another. */

void tagtree_reset(tagtree_t *tree);
/* Reset the tag tree state. */

void tagtree_setvalue(tagtree_t *tree, tagtreenode_t *leaf, int value);
/* Set the value associated with a particular leaf node of a tag tree. */

tagtreenode_t *tagtree_getleaf(tagtree_t *tree, int n);
/* Get a pointer to a particular leaf node. */

int tagtree_decode(tagtree_t *tree, tagtreenode_t *leaf, int threshold,
  bitstream_t *in);
/* Invoke the tag tree decoding procedure. */

int tagtree_encode(tagtree_t *tree, tagtreenode_t *leaf, int threshold,
  bitstream_t *out);
/* Invoke the tag tree encoding procedure. */

void tagtree_dump(tagtree_t *tree, FILE *out);
/* Dump a tag tree (for debugging purposes). */

#endif
