/*
 * Copyright (c) 1999-2000 Image Power, Inc. and the University of
 *   British Columbia.
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * Sun Rasterfile Library
 *
 * $Id$
 */

/******************************************************************************\
* Sun Rasterfile
\******************************************************************************/

#ifndef	RAS_H
#define	RAS_H

/******************************************************************************\
* Includes.
\******************************************************************************/

#include "stream.h"
#include "image.h"

/******************************************************************************\
* Constants.
\******************************************************************************/

#define	RAS_MAGIC	0x59a66a95

#define	RAS_TYPE_OLD	0
#define	RAS_TYPE_STD	1
#define	RAS_TYPE_RLE	2

#define	RAS_MT_NONE	0
#define	RAS_MT_EQUALRGB	1
#define	RAS_MT_RAW	2

/******************************************************************************\
* Types.
\******************************************************************************/

/* Sun Rasterfile header. */
typedef struct {

	uint_fast32_t magic;
	/* signature */

	uint_fast32_t width;
	/* width of image (in pixels) */

	uint_fast32_t height;
	/* height of image (in pixels) */

	uint_fast32_t depth;
	/* number of bits per pixel */

	uint_fast32_t length;
	/* length of image data (in bytes) */

	uint_fast32_t type;
	/* format of image data */

	uint_fast32_t maptype;
	/* colormap type */

	uint_fast32_t maplength;
	/* length of colormap data (in bytes) */

} ras_hdr_t;

/* Color map. */
typedef struct {

	int len;
	/* The number of entries in the color map. */

	int data[256];
	/* The color map data. */

} ras_cmap_t;

/******************************************************************************\
* Macros.
\******************************************************************************/

#define RAS_GETBLUE(x)	(((x) >> 16) & 0xff)
#define	RAS_GETGREEN(x)	(((x) >> 8) & 0xff)
#define RAS_GETRED(x)	((x) & 0xff)

#define RAS_BLUE(x)	(((x) & 0xff) << 16)
#define RAS_GREEN(x)	(((x) & 0xff) << 8)
#define	RAS_RED(x)	((x) & 0xff)

#define RAS_ROWSIZE(hdr) \
	((((hdr)->width * (hdr)->depth + 15) / 16) * 2)
#define RAS_ISRGB(hdr) \
	((hdr)->depth == 24 || (hdr)->depth == 32)

#define	RAS_ONES(n) \
	(((n) == 32) ? 0xffffffffU : ((1 << (n)) - 1))

#endif
